#ifndef __TEST_TYPE_2
#define __TEST_TYPE_2

#include <utility>

class TestType2;

void TT2_create(TestType2 & r);
void TT2_destroy(TestType2 & r);
void TT2_createcopy(TestType2 & r, const TestType2 & a);
void TT2_copy(TestType2 & r, const TestType2 & a);
void setZero(TestType2 & r);
void setOne(TestType2 & r);
void TT2_add(TestType2 & r, const TestType2 & a, const TestType2 & b);
void TT2_sub(TestType2 & r, const TestType2 & a, const TestType2 & b);
void TT2_mul(TestType2 & r, const TestType2 & a, const TestType2 & b);
void TT2_div(TestType2 & r, const TestType2 & a, const TestType2 & b);
void TT2_mod(TestType2 & r, const TestType2 & a, const TestType2 & b);
void TT2_neg(TestType2 & r, const TestType2 & a);

class TestWrapper2
{
private:
    const TestType2 & d_val;
    
public:
    inline TestWrapper2(const TestType2 & val)
        : d_val(val)
    {
    }
    
    inline const TestType2 & evaluate() const
    {
        return d_val;
    }
    
    inline void evalTo(TestType2 & dest)
    {
        TT2_copy(dest, d_val);
    }
};

template<class Op, class Data>
class TestExpression2;

class TestType2
{
private:
    void * d_data;
    
    friend void TT2_create(TestType2 & r);
    friend void TT2_destroy(TestType2 & r);
    friend void TT2_createcopy(TestType2 & r, const TestType2 & a);
    friend void TT2_copy(TestType2 & r, const TestType2 & a);
    friend void setZero(TestType2 & r);
    friend void setOne(TestType2 & r);
    friend void TT2_add(TestType2 & r, const TestType2 & a, const TestType2 & b);
    friend void TT2_sub(TestType2 & r, const TestType2 & a, const TestType2 & b);
    friend void TT2_mul(TestType2 & r, const TestType2 & a, const TestType2 & b);
    friend void TT2_div(TestType2 & r, const TestType2 & a, const TestType2 & b);
    friend void TT2_mod(TestType2 & r, const TestType2 & a, const TestType2 & b);
    friend void TT2_neg(TestType2 & r, const TestType2 & a);
    
public:
    inline TestType2()
    {
        TT2_create(*this);
    }

    inline TestType2(const TestType2 & src)
    {
        TT2_createcopy(*this, src);
    }

    template<class O, class D>
    inline TestType2(const TestExpression2<O, D> & src)
    {
        TT2_createcopy(*this, src.evaluate());
    }

    inline ~TestType2()
    {
        TT2_destroy(*this);
    }

    inline TestType2 & operator = (const TestType2 & src)
    {
        TT2_copy(*this, src);
        return *this;
    }
    
    template<class O, class D>
    inline TestType2 & operator = (const TestExpression2<O, D> & e)
    {
        e.evalTo(*this);
        return *this;
    }
    
    inline TestType2 & operator += (const TestType2 & b)
    {
        TT2_add(*this, *this, b);
        return *this;
    }
    
    inline TestType2 & operator -= (const TestType2 & b)
    {
        TT2_sub(*this, *this, b);
        return *this;
    }
    
    inline TestType2 & operator *= (const TestType2 & b)
    {
        TT2_mul(*this, *this, b);
        return *this;
    }
    
    inline TestType2 & operator /= (const TestType2 & b)
    {
        TT2_div(*this, *this, b);
        return *this;
    }
    
    inline TestType2 & operator %= (const TestType2 & b)
    {
        TT2_mod(*this, *this, b);
        return *this;
    }
    
    template<class O, class D>
    inline TestType2 & operator += (const TestExpression2<O, D> & e)
    {
        TT2_add(*this, *this, e.evaluate());
        return *this;
    }
    
    template<class O, class D>
    inline TestType2 & operator -= (const TestExpression2<O, D> & e)
    {
        TT2_sub(*this, *this, e.evaluate());
        return *this;
    }
    
    template<class O, class D>
    inline TestType2 & operator *= (const TestExpression2<O, D> & e)
    {
        TT2_mul(*this, *this, e.evaluate());
        return *this;
    }
    
    template<class O, class D>
    inline TestType2 & operator /= (const TestExpression2<O, D> & e)
    {
        TT2_div(*this, *this, e.evaluate());
        return *this;
    }
    
    template<class O, class D>
    inline TestType2 & operator %= (const TestExpression2<O, D> & e)
    {
        TT2_mod(*this, *this, e.evaluate());
        return *this;
    }
};

template<class Op, class Data>
class TestExpression2
{
private:
    Op d_op;
    Data d_data;
    
public:
    inline TestExpression2(const Op & op, const Data & data)
        : d_op(op), d_data(data)
    {
    }
    
    operator TestType2 () const
    {
        TestType2 res;
        evalTo(res);
        return res;
    }
    
    inline TestType2 evaluate() const
    {
        TestType2 res;
        evalTo(res);
        return res;
    }
    
    inline void evalTo(TestType2 & dest) const
    {
        d_op.evalTo(dest, d_data);
    }
};

class AddOp2
{
public:
    template<class A, class B>
    inline void evalTo(TestType2 & dest, const std::pair<A, B> & data) const
    {
        TT2_add(dest, data.first.evaluate(), data.second.evaluate());
    }
};

class SubOp2
{
public:
    template<class A, class B>
    inline void evalTo(TestType2 & dest, const std::pair<A, B> & data) const
    {
        TT2_sub(dest, data.first.evaluate(), data.second.evaluate());
    }
};

class MulOp2
{
public:
    template<class A, class B>
    inline void evalTo(TestType2 & dest, const std::pair<A, B> & data) const
    {
        TT2_mul(dest, data.first.evaluate(), data.second.evaluate());
    }
};

class DivOp2
{
public:
    template<class A, class B>
    inline void evalTo(TestType2 & dest, const std::pair<A, B> & data) const
    {
        TT2_div(dest, data.first.evaluate(), data.second.evaluate());
    }
};

class ModOp2
{
public:
    template<class A, class B>
    inline void evalTo(TestType2 & dest, const std::pair<A, B> & data) const
    {
        TT2_mod(dest, data.first.evaluate(), data.second.evaluate());
    }
};

class NegOp2
{
public:
    template<class A>
    inline void evalTo(TestType2 & dest, const A & data) const
    {
        TT2_neg(dest, data.evaluate());
    }
};

inline TestExpression2<AddOp2, std::pair<TestWrapper2, TestWrapper2> > operator + (const TestType2 & a, const TestType2 & b)
{ return TestExpression2<AddOp2, std::pair<TestWrapper2, TestWrapper2> >(AddOp2(), std::make_pair(TestWrapper2(a), TestWrapper2(b))); }
inline TestExpression2<SubOp2, std::pair<TestWrapper2, TestWrapper2> > operator - (const TestType2 & a, const TestType2 & b)
{ return TestExpression2<SubOp2, std::pair<TestWrapper2, TestWrapper2> >(SubOp2(), std::make_pair(TestWrapper2(a), TestWrapper2(b))); }
inline TestExpression2<MulOp2, std::pair<TestWrapper2, TestWrapper2> > operator * (const TestType2 & a, const TestType2 & b)
{ return TestExpression2<MulOp2, std::pair<TestWrapper2, TestWrapper2> >(MulOp2(), std::make_pair(TestWrapper2(a), TestWrapper2(b))); }
inline TestExpression2<DivOp2, std::pair<TestWrapper2, TestWrapper2> > operator / (const TestType2 & a, const TestType2 & b)
{ return TestExpression2<DivOp2, std::pair<TestWrapper2, TestWrapper2> >(DivOp2(), std::make_pair(TestWrapper2(a), TestWrapper2(b))); }
inline TestExpression2<ModOp2, std::pair<TestWrapper2, TestWrapper2> > operator % (const TestType2 & a, const TestType2 & b)
{ return TestExpression2<ModOp2, std::pair<TestWrapper2, TestWrapper2> >(ModOp2(), std::make_pair(TestWrapper2(a), TestWrapper2(b))); }

template<class O2, class D2>
inline TestExpression2<AddOp2, std::pair<TestWrapper2, TestExpression2<O2, D2> > > operator + (const TestType2 & a, const TestExpression2<O2, D2> & b)
{ return TestExpression2<AddOp2, std::pair<TestWrapper2, TestExpression2<O2, D2> > >(AddOp2(), std::make_pair(TestWrapper2(a), b)); }
template<class O2, class D2>
inline TestExpression2<SubOp2, std::pair<TestWrapper2, TestExpression2<O2, D2> > > operator - (const TestType2 & a, const TestExpression2<O2, D2> & b)
{ return TestExpression2<SubOp2, std::pair<TestWrapper2, TestExpression2<O2, D2> > >(SubOp2(), std::make_pair(TestWrapper2(a), b)); }
template<class O2, class D2>
inline TestExpression2<MulOp2, std::pair<TestWrapper2, TestExpression2<O2, D2> > > operator * (const TestType2 & a, const TestExpression2<O2, D2> & b)
{ return TestExpression2<MulOp2, std::pair<TestWrapper2, TestExpression2<O2, D2> > >(MulOp2(), std::make_pair(TestWrapper2(a), b)); }
template<class O2, class D2>
inline TestExpression2<DivOp2, std::pair<TestWrapper2, TestExpression2<O2, D2> > > operator / (const TestType2 & a, const TestExpression2<O2, D2> & b)
{ return TestExpression2<DivOp2, std::pair<TestWrapper2, TestExpression2<O2, D2> > >(DivOp2(), std::make_pair(TestWrapper2(a), b)); }
template<class O2, class D2>
inline TestExpression2<ModOp2, std::pair<TestWrapper2, TestExpression2<O2, D2> > > operator % (const TestType2 & a, const TestExpression2<O2, D2> & b)
{ return TestExpression2<ModOp2, std::pair<TestWrapper2, TestExpression2<O2, D2> > >(ModOp2(), std::make_pair(TestWrapper2(a), b)); }

template<class O1, class D1>
inline TestExpression2<AddOp2, std::pair<TestExpression2<O1, D1>, TestWrapper2> > operator + (const TestExpression2<O1, D1> & a, const TestType2 & b)
{ return TestExpression2<AddOp2, std::pair<TestExpression2<O1, D1>, TestWrapper2> >(AddOp2(), std::make_pair(a, TestWrapper2(b))); }
template<class O1, class D1>
inline TestExpression2<SubOp2, std::pair<TestExpression2<O1, D1>, TestWrapper2> > operator - (const TestExpression2<O1, D1> & a, const TestType2 & b)
{ return TestExpression2<SubOp2, std::pair<TestExpression2<O1, D1>, TestWrapper2> >(SubOp2(), std::make_pair(a, TestWrapper2(b))); }
template<class O1, class D1>
inline TestExpression2<MulOp2, std::pair<TestExpression2<O1, D1>, TestWrapper2> > operator * (const TestExpression2<O1, D1> & a, const TestType2 & b)
{ return TestExpression2<MulOp2, std::pair<TestExpression2<O1, D1>, TestWrapper2> >(MulOp2(), std::make_pair(a, TestWrapper2(b))); }
template<class O1, class D1>
inline TestExpression2<DivOp2, std::pair<TestExpression2<O1, D1>, TestWrapper2> > operator / (const TestExpression2<O1, D1> & a, const TestType2 & b)
{ return TestExpression2<DivOp2, std::pair<TestExpression2<O1, D1>, TestWrapper2> >(DivOp2(), std::make_pair(a, TestWrapper2(b))); }
template<class O1, class D1>
inline TestExpression2<ModOp2, std::pair<TestExpression2<O1, D1>, TestWrapper2> > operator % (const TestExpression2<O1, D1> & a, const TestType2 & b)
{ return TestExpression2<ModOp2, std::pair<TestExpression2<O1, D1>, TestWrapper2> >(ModOp2(), std::make_pair(a, TestWrapper2(b))); }

template<class O1, class D1, class O2, class D2>
inline TestExpression2<AddOp2, std::pair<TestExpression2<O1, D1>, TestExpression2<O2, D2> > > operator + (const TestExpression2<O1, D1> & a, const TestExpression2<O2, D2> & b)
{ return TestExpression2<AddOp2, std::pair<TestExpression2<O1, D1>, TestExpression2<O2, D2> > >(AddOp2(), std::make_pair(a, b)); }
template<class O1, class D1, class O2, class D2>
inline TestExpression2<SubOp2, std::pair<TestExpression2<O1, D1>, TestExpression2<O2, D2> > > operator - (const TestExpression2<O1, D1> & a, const TestExpression2<O2, D2> & b)
{ return TestExpression2<SubOp2, std::pair<TestExpression2<O1, D1>, TestExpression2<O2, D2> > >(SubOp2(), std::make_pair(a, b)); }
template<class O1, class D1, class O2, class D2>
inline TestExpression2<MulOp2, std::pair<TestExpression2<O1, D1>, TestExpression2<O2, D2> > > operator * (const TestExpression2<O1, D1> & a, const TestExpression2<O2, D2> & b)
{ return TestExpression2<MulOp2, std::pair<TestExpression2<O1, D1>, TestExpression2<O2, D2> > >(MulOp2(), std::make_pair(a, b)); }
template<class O1, class D1, class O2, class D2>
inline TestExpression2<DivOp2, std::pair<TestExpression2<O1, D1>, TestExpression2<O2, D2> > > operator / (const TestExpression2<O1, D1> & a, const TestExpression2<O2, D2> & b)
{ return TestExpression2<DivOp2, std::pair<TestExpression2<O1, D1>, TestExpression2<O2, D2> > >(DivOp2(), std::make_pair(a, b)); }
template<class O1, class D1, class O2, class D2>
inline TestExpression2<ModOp2, std::pair<TestExpression2<O1, D1>, TestExpression2<O2, D2> > > operator % (const TestExpression2<O1, D1> & a, const TestExpression2<O2, D2> & b)
{ return TestExpression2<ModOp2, std::pair<TestExpression2<O1, D1>, TestExpression2<O2, D2> > >(ModOp2(), std::make_pair(a, b)); }

inline TestExpression2<NegOp2, TestWrapper2> operator - (const TestType2 & a)
{ return TestExpression2<NegOp2, TestWrapper2>(NegOp2(), TestWrapper2(a)); }

template<class O1, class D1>
inline TestExpression2<NegOp2, TestExpression2<O1, D1> > operator - (const TestExpression2<O1, D1> & a)
{ return TestExpression2<NegOp2, TestExpression2<O1, D1> >(NegOp2(), a); }

#endif
